#ifndef ELEMENTS_CYLINDRICAL_3_HH
#define ELEMENTS_CYLINDRICAL_3_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void refractCylindrical3(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float n1, float n2)
{
  poly.clear();
  //=================================================== 
  // taylor series degree 3 of refracted position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term4f( 1 ,  1 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term4f( (n2-n1)*pow(n2,-1)*pow(R,-2)/2 ,  3 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 0 ].add(Term4f( (n2-n1)*pow(n2,-1)*pow(R,-1)/2 ,  2 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term4f( 1 ,  0 ,  1 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 1 ].add(Term4f( (n2-n1)*pow(n2,-1)*pow(R,-1)/2 ,  2 ,  0 ,  0 ,  1 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term4f( -(n2-n1)*pow(n2,-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(pow(n2,3)-2*n1*pow(n2,2)+2*pow(n1,2)*n2-pow(n1,3))*pow(R,-3)/2 ,  3 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term4f( n1*pow(n2,-1) ,  0 ,  0 ,  1 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(pow(n2,3)-2*n1*pow(n2,2)+4*pow(n1,2)*n2-3*pow(n1,3))*pow(R,-2)/2 ,  2 ,  0 ,  1 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)+2*pow(n1,2)*n2-3*pow(n1,3))*pow(R,-1)/2 ,  1 ,  0 ,  2 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  3 ,  0 )); 
  // dy^2*x  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))*pow(R,-1)/2 ,  1 ,  0 ,  0 ,  2 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  1 ,  2 )); 
   
  //=================================================== 
  // taylor series degree 3 of refracted direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term4f( n1*pow(n2,-1) ,  0 ,  0 ,  0 ,  1 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term4f( pow(n2,-3)*(n1*pow(n2,2)-2*pow(n1,2)*n2+pow(n1,3))*pow(R,-2)/2 ,  2 ,  0 ,  0 ,  1 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term4f( -(pow(n1,2)*n2-pow(n1,3))*pow(n2,-3)*pow(R,-1) ,  1 ,  0 ,  1 ,  1 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  2 ,  1 )); 
  // dy^3  :  
  poly[ 3 ].add(Term4f( -pow(n2,-3)*(n1*pow(n2,2)-pow(n1,3))/2 ,  0 ,  0 ,  0 ,  3 )); 
}


template<typename scalar, int numVarsIn, int numVarsOut>
static inline void reflectCylindrical3(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float n1, float n2)
{
  poly.clear();
  //=================================================== 
  // taylor series degree 3 of reflected position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term4f( 1 ,  1 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term4f( -pow(R,-2) ,  3 ,  0 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term4f( 1.0f ,  0 ,  1 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term4f( 2*pow(R,-1) ,  1 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term4f( 3*pow(R,-3) ,  3 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term4f( 1 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term4f( 5*pow(R,-2) ,  2 ,  0 ,  1 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term4f( 2*pow(R,-1) ,  1 ,  0 ,  2 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term4f( 1 ,  0 ,  0 ,  0 ,  1 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term4f( 2*pow(R,-2) ,  2 ,  0 ,  0 ,  1 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term4f( 2*pow(R,-1) ,  1 ,  0 ,  1 ,  1 )); 

}

} // namespace poly

#endif
