#ifndef ELEMENTS_CYLINDRICAL_SPECTRAL_HH
#define ELEMENTS_CYLINDRICAL_SPECTRAL_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void reflectCylindricalSpectralConcave(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
{
  poly.clear();
  if(numVarsIn > 4)
    poly[4].unity(4);

  //=================================================== 
  // taylor series degree 3 of reflected position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term5f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term5f( -pow(R,-2) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term5f( 1.0 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term5f( 3*pow(R,-3) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( 1 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term5f( 5*pow(R,-2) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term5f( 1 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term5f( 2*pow(R,-2) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  1 ,  1 ,  0 ));
}

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void refractCylindricalSpectralConcave(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
{
  poly.clear();
  if(numVarsIn > 4)
    poly[4].unity(4);
  //=================================================== 
  // taylor series degree 3 of transmitted position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term5f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-2) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 0 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-1) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term5f( 1 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 1 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-1) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(48*A2-64*B1)*pow(B2,2)+(-48*A2*B1+12*pow(A2,2)+12*A1*A2)*B2-24*pow(A2,2)*B1+pow(A2,3)-A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(-128*B1+48*A2-32*A1)*pow(B2,2)+(128*pow(B1,2)+(64*A1-64*A2)*B1+12*pow(A2,2)-16*A1*A2+8*pow(A1,2))*B2-64*pow(B1,3)+(32*A2-48*A1)*pow(B1,2)+(-8*pow(A2,2)+16*A1*A2-12*pow(A1,2))*B1+pow(A2,3)-2*A1*pow(A2,2)+2*pow(A1,2)*A2-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-3) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( (64*B1*pow(B2,2)+(48*A2*B1-12*A1*A2)*B2+24*pow(A2,2)*B1+A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(-128*B1+48*A2-32*A1)*pow(B2,2)+(256*pow(B1,2)+(128*A1-64*A2)*B1+12*pow(A2,2)-16*A1*A2+16*pow(A1,2))*B2-192*pow(B1,3)+(64*A2-144*A1)*pow(B1,2)+(-8*pow(A2,2)+32*A1*A2-36*pow(A1,2))*B1+pow(A2,3)-2*A1*pow(A2,2)+4*pow(A1,2)*A2-3*pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-2) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(128*pow(B1,2)+(32*A2+64*A1)*B1+8*A1*A2+8*pow(A1,2))*B2-192*pow(B1,3)+(32*A2-144*A1)*pow(B1,2)+(4*pow(A2,2)+16*A1*A2-36*pow(A1,2))*B1+A1*pow(A2,2)+2*pow(A1,2)*A2-3*pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy^2*x  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // x*lambda  :  
  poly[ 2 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*lambda  :  
  poly[ 2 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  1 ,  0 ,  1 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term5f( (64*B1*pow(B2,2)+(48*A2*B1-12*A1*A2)*B2+24*pow(A2,2)*B1+A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term5f( ((64*B1+16*A1)*pow(B2,2)+(-128*pow(B1,2)+(32*A2-64*A1)*B1+8*A1*A2-8*pow(A1,2))*B2+64*pow(B1,3)+(48*A1-32*A2)*pow(B1,2)+(4*pow(A2,2)-16*A1*A2+12*pow(A1,2))*B1+A1*pow(A2,2)-2*pow(A1,2)*A2+pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-2) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term5f( -((64*pow(B1,2)+32*A1*B1+4*pow(A1,2))*B2-64*pow(B1,3)+(16*A2-48*A1)*pow(B1,2)+(8*A1*A2-12*pow(A1,2))*B1+pow(A1,2)*A2-pow(A1,3))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^3  :  
  poly[ 3 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
  // dy*lambda  :  
  poly[ 3 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  0 ,  1 ,  1 )); 
}

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void refractCylindricalSpectral(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
{
  if(R < 0)
  {
    refractCylindricalSpectralConcave(poly, R, A1, B1, A2, B2);
    return;
  }
  poly.clear();
  if(numVarsIn > 4)
    poly[4].unity(4);

  //=================================================== 
  // taylor series degree 3 of transmitted position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term5f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-2) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 0 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-1) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted position y 
  //=================================================== 
  // y  :  
  poly[ 1 ].add(Term5f( 1 ,  0 ,  1 ,  0 ,  0 ,  0 )); 
  // dy*x^2  :  
  poly[ 1 ].add(Term5f( (4*B2-4*B1+A2-A1)*pow(8*B2+2*A2,-1)*pow(R,-1) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(48*A2-64*B1)*pow(B2,2)+(-48*A2*B1+12*pow(A2,2)+12*A1*A2)*B2-24*pow(A2,2)*B1+pow(A2,3)-A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(-128*B1+48*A2-32*A1)*pow(B2,2)+(128*pow(B1,2)+(64*A1-64*A2)*B1+12*pow(A2,2)-16*A1*A2+8*pow(A1,2))*B2-64*pow(B1,3)+(32*A2-48*A1)*pow(B1,2)+(-8*pow(A2,2)+16*A1*A2-12*pow(A1,2))*B1+pow(A2,3)-2*A1*pow(A2,2)+2*pow(A1,2)*A2-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-3) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( (64*B1*pow(B2,2)+(48*A2*B1-12*A1*A2)*B2+24*pow(A2,2)*B1+A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term5f( -(64*pow(B2,3)+(-128*B1+48*A2-32*A1)*pow(B2,2)+(256*pow(B1,2)+(128*A1-64*A2)*B1+12*pow(A2,2)-16*A1*A2+16*pow(A1,2))*B2-192*pow(B1,3)+(64*A2-144*A1)*pow(B1,2)+(-8*pow(A2,2)+32*A1*A2-36*pow(A1,2))*B1+pow(A2,3)-2*A1*pow(A2,2)+4*pow(A1,2)*A2-3*pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-2) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(128*pow(B1,2)+(32*A2+64*A1)*B1+8*A1*A2+8*pow(A1,2))*B2-192*pow(B1,3)+(32*A2-144*A1)*pow(B1,2)+(4*pow(A2,2)+16*A1*A2-36*pow(A1,2))*B1+A1*pow(A2,2)+2*pow(A1,2)*A2-3*pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
  // dx^3  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  3 ,  0 ,  0 )); 
  // dy^2*x  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  2 ,  0 )); 
  // dx*dy^2  :  
  poly[ 2 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  1 ,  2 ,  0 )); 
  // x*lambda  :  
  poly[ 2 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  1 )); 
  // dx*lambda  :  
  poly[ 2 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  1 ,  0 ,  1 )); 
   
  //=================================================== 
  // taylor series degree 3 of transmitted direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term5f( (64*B1*pow(B2,2)+(48*A2*B1-12*A1*A2)*B2+24*pow(A2,2)*B1+A1*pow(A2,2))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term5f( ((64*B1+16*A1)*pow(B2,2)+(-128*pow(B1,2)+(32*A2-64*A1)*B1+8*A1*A2-8*pow(A1,2))*B2+64*pow(B1,3)+(48*A1-32*A2)*pow(B1,2)+(4*pow(A2,2)-16*A1*A2+12*pow(A1,2))*B1+A1*pow(A2,2)-2*pow(A1,2)*A2+pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1)*pow(R,-2) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term5f( -((64*pow(B1,2)+32*A1*B1+4*pow(A1,2))*B2-64*pow(B1,3)+(16*A2-48*A1)*pow(B1,2)+(8*A1*A2-12*pow(A1,2))*B1+pow(A1,2)*A2-pow(A1,3))*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1)*pow(R,-1) ,  1 ,  0 ,  1 ,  1 ,  0 )); 
  // dx^2*dy  :  
  poly[ 3 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  2 ,  1 ,  0 )); 
  // dy^3  :  
  poly[ 3 ].add(Term5f( -((64*B1+16*A1)*pow(B2,2)+(32*A2*B1+8*A1*A2)*B2-64*pow(B1,3)-48*A1*pow(B1,2)+(4*pow(A2,2)-12*pow(A1,2))*B1+A1*pow(A2,2)-pow(A1,3))*pow(128*pow(B2,3)+96*A2*pow(B2,2)+24*pow(A2,2)*B2+2*pow(A2,3),-1) ,  0 ,  0 ,  0 ,  3 ,  0 )); 
  // dy*lambda  :  
  poly[ 3 ].add(Term5f( (64*A1*A2*B2-64*pow(A2,2)*B1)*pow(64*pow(B2,3)+48*A2*pow(B2,2)+12*pow(A2,2)*B2+pow(A2,3),-1) ,  0 ,  0 ,  0 ,  1 ,  1 ));
}


template<typename scalar, int numVarsIn, int numVarsOut>
static inline void reflectCylindricalSpectral(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly, float R, float A1, float B1, float A2, float B2)
{
  if(R < 0)
  {
    reflectCylindricalSpectralConcave(poly, R, A1, B1, A2, B2);
    return;
  }
  poly.clear();
  if(numVarsIn > 4)
    poly[4].unity(4);

  //=================================================== 
  // taylor series degree 3 of reflected position x 
  //=================================================== 
  // x  :  
  poly[ 0 ].add(Term5f( 1 ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 0 ].add(Term5f( -pow(R,-2) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected position y 
  //=================================================== 
  // 1  :  
  poly[ 1 ].add(Term5f( 1.0f ,  0 ,  1 ,  0 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dx 
  //=================================================== 
  // x  :  
  poly[ 2 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  0 ,  0 ,  0 )); 
  // x^3  :  
  poly[ 2 ].add(Term5f( 3*pow(R,-3) ,  3 ,  0 ,  0 ,  0 ,  0 )); 
  // dx  :  
  poly[ 2 ].add(Term5f( 1 ,  0 ,  0 ,  1 ,  0 ,  0 )); 
  // dx*x^2  :  
  poly[ 2 ].add(Term5f( 5*pow(R,-2) ,  2 ,  0 ,  1 ,  0 ,  0 )); 
  // dx^2*x  :  
  poly[ 2 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  2 ,  0 ,  0 )); 
   
  //=================================================== 
  // taylor series degree 3 of reflected direction dy 
  //=================================================== 
  // dy  :  
  poly[ 3 ].add(Term5f( 1 ,  0 ,  0 ,  0 ,  1 ,  0 )); 
  // dy*x^2  :  
  poly[ 3 ].add(Term5f( 2*pow(R,-2) ,  2 ,  0 ,  0 ,  1 ,  0 )); 
  // dx*dy*x  :  
  poly[ 3 ].add(Term5f( 2*pow(R,-1) ,  1 ,  0 ,  1 ,  1 ,  0 )); 

}

} // namespace poly

#endif
