#ifndef ELEMENTS_PATHSPACE_PROPAGATEX_HH
#define ELEMENTS_PATHSPACE_PROPAGATEX_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

// surface reflection parameterized over three vertices x0 x1 x2 in their respective tangent spaces.
// the only output is f (measurement contribution)
template<typename scalar, int numVars>
static inline void surfacex_propagate(
    truncPoly<scalar, numVars> &poly,
    float *T0,             // 3x3 tangent space [a,b,n] of v0, z is up
    float *T1,             // v1
    float *v0,             // vertex position at v0
    float *v1)             // vertex position at v1
{
  assert(numVars >= 4); // [x0,y0,x1,y1]
  poly.terms.clear();

  // init tangent space [abn][xyz][01]
  const float vx0 = v0[0], vy0 = v0[1], vz0 = v0[2];
  const float vx1 = v1[0], vy1 = v1[1], vz1 = v1[2];

  const float ax0 = T0[3*0 + 0];
  const float bx0 = T0[3*0 + 1];
  const float nx0 = T0[3*0 + 2];
  const float ay0 = T0[3*1 + 0];
  const float by0 = T0[3*1 + 1];
  const float ny0 = T0[3*1 + 2];
  // const float az0 = T0[3*2 + 0]; // actually unused in the expressions below
  // const float bz0 = T0[3*2 + 1];
  const float nz0 = T0[3*2 + 2];

  const float ax1 = T1[3*0 + 0];
  const float bx1 = T1[3*0 + 1];
  const float nx1 = T1[3*0 + 2];
  const float ay1 = T1[3*1 + 0];
  const float by1 = T1[3*1 + 1];
  const float ny1 = T1[3*1 + 2];
  // const float az1 = T1[3*2 + 0];
  // const float bz1 = T1[3*2 + 1];
  // const float nz1 = T1[3*2 + 2];

// 1  :  
poly.add(Term4f( (nz0*vz1-nz0*vz0+ny0*vy1-ny0*vy0+nx0*vx1-nx0*vx0)*pow(pow(vz1,2)-2*vz0*vz1+pow(vz0,2)+pow(vy1,2)-2*vy0*vy1+pow(vy0,2)+pow(vx1,2)-2*vx0*vx1+pow(vx0,2),-3/2) ,  0 ,  0 ,  0 ,  0 )); 
// x0  :  
poly.add(Term4f( ((2*nx0*nz0-bx0*ny0-ax0*nx0)*pow(vz1,2)+((-4*nx0*nz0+2*bx0*ny0+2*ax0*nx0)*vz0+(3*bx0*nz0+3*nx0*ny0)*vy1+(-3*bx0*nz0-3*nx0*ny0)*vy0+(3*ax0*nz0+3*pow(nx0,2))*vx1+(-3*ax0*nz0-3*pow(nx0,2))*vx0)*vz1+(2*nx0*nz0-bx0*ny0-ax0*nx0)*pow(vz0,2)+((-3*bx0*nz0-3*nx0*ny0)*vy1+(3*bx0*nz0+3*nx0*ny0)*vy0+(-3*ax0*nz0-3*pow(nx0,2))*vx1+(3*ax0*nz0+3*pow(nx0,2))*vx0)*vz0+(-nx0*nz0+2*bx0*ny0-ax0*nx0)*pow(vy1,2)+((2*nx0*nz0-4*bx0*ny0+2*ax0*nx0)*vy0+(3*ax0*ny0+3*bx0*nx0)*vx1+(-3*ax0*ny0-3*bx0*nx0)*vx0)*vy1+(-nx0*nz0+2*bx0*ny0-ax0*nx0)*pow(vy0,2)+((-3*ax0*ny0-3*bx0*nx0)*vx1+(3*ax0*ny0+3*bx0*nx0)*vx0)*vy0+(-nx0*nz0-bx0*ny0+2*ax0*nx0)*pow(vx1,2)+(2*nx0*nz0+2*bx0*ny0-4*ax0*nx0)*vx0*vx1+(-nx0*nz0-bx0*ny0+2*ax0*nx0)*pow(vx0,2))*pow(pow(vz1,2)-2*vz0*vz1+pow(vz0,2)+pow(vy1,2)-2*vy0*vy1+pow(vy0,2)+pow(vx1,2)-2*vx0*vx1+pow(vx0,2),-1/2)*pow(pow(vz1,4)-4*vz0*pow(vz1,3)+(6*pow(vz0,2)+2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz1,2)+((-4*pow(vy1,2)+8*vy0*vy1-4*pow(vy0,2)-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vz0-4*pow(vz0,3))*vz1+pow(vz0,4)+(2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz0,2)+pow(vy1,4)-4*vy0*pow(vy1,3)+(6*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy1,2)+((-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vy0-4*pow(vy0,3))*vy1+pow(vy0,4)+(2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy0,2)+pow(vx1,4)-4*vx0*pow(vx1,3)+6*pow(vx0,2)*pow(vx1,2)-4*pow(vx0,3)*vx1+pow(vx0,4),-1) ,  1 ,  0 ,  0 ,  0 )); 
// y0  :  
poly.add(Term4f( ((2*ny0*nz0-by0*ny0-ay0*nx0)*pow(vz1,2)+((-4*ny0*nz0+2*by0*ny0+2*ay0*nx0)*vz0+(3*by0*nz0+3*pow(ny0,2))*vy1+(-3*by0*nz0-3*pow(ny0,2))*vy0+(3*ay0*nz0+3*nx0*ny0)*vx1+(-3*ay0*nz0-3*nx0*ny0)*vx0)*vz1+(2*ny0*nz0-by0*ny0-ay0*nx0)*pow(vz0,2)+((-3*by0*nz0-3*pow(ny0,2))*vy1+(3*by0*nz0+3*pow(ny0,2))*vy0+(-3*ay0*nz0-3*nx0*ny0)*vx1+(3*ay0*nz0+3*nx0*ny0)*vx0)*vz0+(-ny0*nz0+2*by0*ny0-ay0*nx0)*pow(vy1,2)+((2*ny0*nz0-4*by0*ny0+2*ay0*nx0)*vy0+(3*ay0*ny0+3*by0*nx0)*vx1+(-3*ay0*ny0-3*by0*nx0)*vx0)*vy1+(-ny0*nz0+2*by0*ny0-ay0*nx0)*pow(vy0,2)+((-3*ay0*ny0-3*by0*nx0)*vx1+(3*ay0*ny0+3*by0*nx0)*vx0)*vy0+(-ny0*nz0-by0*ny0+2*ay0*nx0)*pow(vx1,2)+(2*ny0*nz0+2*by0*ny0-4*ay0*nx0)*vx0*vx1+(-ny0*nz0-by0*ny0+2*ay0*nx0)*pow(vx0,2))*pow(pow(vz1,2)-2*vz0*vz1+pow(vz0,2)+pow(vy1,2)-2*vy0*vy1+pow(vy0,2)+pow(vx1,2)-2*vx0*vx1+pow(vx0,2),-1/2)*pow(pow(vz1,4)-4*vz0*pow(vz1,3)+(6*pow(vz0,2)+2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz1,2)+((-4*pow(vy1,2)+8*vy0*vy1-4*pow(vy0,2)-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vz0-4*pow(vz0,3))*vz1+pow(vz0,4)+(2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz0,2)+pow(vy1,4)-4*vy0*pow(vy1,3)+(6*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy1,2)+((-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vy0-4*pow(vy0,3))*vy1+pow(vy0,4)+(2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy0,2)+pow(vx1,4)-4*vx0*pow(vx1,3)+6*pow(vx0,2)*pow(vx1,2)-4*pow(vx0,3)*vx1+pow(vx0,4),-1) ,  0 ,  1 ,  0 ,  0 )); 
// x1  :  
poly.add(Term4f( -((2*nx1*nz0-bx1*ny0-ax1*nx0)*pow(vz1,2)+((-4*nx1*nz0+2*bx1*ny0+2*ax1*nx0)*vz0+(3*bx1*nz0+3*nx1*ny0)*vy1+(-3*bx1*nz0-3*nx1*ny0)*vy0+(3*ax1*nz0+3*nx0*nx1)*vx1+(-3*ax1*nz0-3*nx0*nx1)*vx0)*vz1+(2*nx1*nz0-bx1*ny0-ax1*nx0)*pow(vz0,2)+((-3*bx1*nz0-3*nx1*ny0)*vy1+(3*bx1*nz0+3*nx1*ny0)*vy0+(-3*ax1*nz0-3*nx0*nx1)*vx1+(3*ax1*nz0+3*nx0*nx1)*vx0)*vz0+(-nx1*nz0+2*bx1*ny0-ax1*nx0)*pow(vy1,2)+((2*nx1*nz0-4*bx1*ny0+2*ax1*nx0)*vy0+(3*ax1*ny0+3*bx1*nx0)*vx1+(-3*ax1*ny0-3*bx1*nx0)*vx0)*vy1+(-nx1*nz0+2*bx1*ny0-ax1*nx0)*pow(vy0,2)+((-3*ax1*ny0-3*bx1*nx0)*vx1+(3*ax1*ny0+3*bx1*nx0)*vx0)*vy0+(-nx1*nz0-bx1*ny0+2*ax1*nx0)*pow(vx1,2)+(2*nx1*nz0+2*bx1*ny0-4*ax1*nx0)*vx0*vx1+(-nx1*nz0-bx1*ny0+2*ax1*nx0)*pow(vx0,2))*pow(pow(vz1,2)-2*vz0*vz1+pow(vz0,2)+pow(vy1,2)-2*vy0*vy1+pow(vy0,2)+pow(vx1,2)-2*vx0*vx1+pow(vx0,2),-1/2)*pow(pow(vz1,4)-4*vz0*pow(vz1,3)+(6*pow(vz0,2)+2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz1,2)+((-4*pow(vy1,2)+8*vy0*vy1-4*pow(vy0,2)-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vz0-4*pow(vz0,3))*vz1+pow(vz0,4)+(2*pow(vy1,2)-4*vy0*vy1+2*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vz0,2)+pow(vy1,4)-4*vy0*pow(vy1,3)+(6*pow(vy0,2)+2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy1,2)+((-4*pow(vx1,2)+8*vx0*vx1-4*pow(vx0,2))*vy0-4*pow(vy0,3))*vy1+pow(vy0,4)+(2*pow(vx1,2)-4*vx0*vx1+2*pow(vx0,2))*pow(vy0,2)+pow(vx1,4)-4*vx0*pow(vx1,3)+6*pow(vx0,2)*pow(vx1,2)-4*pow(vx0,3)*vx1+pow(vx0,4),-1) ,  0 ,  0 ,  1 ,  0 )); 
// y1  :  
poly.add(Term4f( -((2*ny1*nz0-by1*ny0-ay1*nx0)*pow(vz1,2)+((-4*ny1*nz0+2*by1*ny0+2*ay1*nx0)*vz0+(3*by1*nz0+3*ny0*ny1)*vy1+(-3*by1*nz0-3*ny0*ny1)*vy0+(3*ay1*nz0+3*nx0*ny1)*vx1+(-3*ay1*nz0-3*nx0*ny1)*vx0)*vz1+(2*ny1*nz0-by1*ny0-ay1*nx0)*pow(vz0,2)+((-3*by1*nz0-3*ny0*ny1)*vy1+(3*by1*nz0+3*ny0*ny1)*vy0+(-3*ay1*nz0-3*nx0*ny1)*vx1+(3*ay1*nz0+3*nx0*ny1)*vx0)*vz0+(-ny1*nz0+2*by1*ny0-ay1*nx0)*pow(vy1,2)+((2*ny1*nz0-4*by1*ny0+2*ay1*nx0)*vy0+(3*ay1*ny0+3*by1*nx0)*vx1+(-3*ay1*ny0-3*by1*nx0)*vx0)*vy1+(-ny1*nz0+2*by1*ny0-ay1*nx0)*pow(vy0,2)+((-3*ay1*ny0-3*by1*nx0)*vx1+(3*ay1*ny0+3*by1*nx0)*vx0)*vy0+(-ny1*nz0-by1*ny0+2*ay1*nx0)*pow(vx1,2)+(2*ny1*nz0+2*by1*ny0-4*ay1*nx0)*vx0*vx1+(-ny1*nz0-by1*ny0+2*ay1*nx0)*pow(vx0,2))*pow(pow(vz1,2)-2*vz0*vz1+pow(vz0,2)+pow(vy1,2)-2*vy0*vy1+pow(vy0,2)+pow(vx1,2)-2*vx0*vx1+pow(vx0,2),1/2)*pow(pow(vz1,6)-6*vz0*pow(vz1,5)+(15*pow(vz0,2)+3*pow(vy1,2)-6*vy0*vy1+3*pow(vy0,2)+3*pow(vx1,2)-6*vx0*vx1+3*pow(vx0,2))*pow(vz1,4)+((-12*pow(vy1,2)+24*vy0*vy1-12*pow(vy0,2)-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*vz0-20*pow(vz0,3))*pow(vz1,3)+(15*pow(vz0,4)+(18*pow(vy1,2)-36*vy0*vy1+18*pow(vy0,2)+18*pow(vx1,2)-36*vx0*vx1+18*pow(vx0,2))*pow(vz0,2)+3*pow(vy1,4)-12*vy0*pow(vy1,3)+(18*pow(vy0,2)+6*pow(vx1,2)-12*vx0*vx1+6*pow(vx0,2))*pow(vy1,2)+((-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*vy0-12*pow(vy0,3))*vy1+3*pow(vy0,4)+(6*pow(vx1,2)-12*vx0*vx1+6*pow(vx0,2))*pow(vy0,2)+3*pow(vx1,4)-12*vx0*pow(vx1,3)+18*pow(vx0,2)*pow(vx1,2)-12*pow(vx0,3)*vx1+3*pow(vx0,4))*pow(vz1,2)+(-6*pow(vz0,5)+(-12*pow(vy1,2)+24*vy0*vy1-12*pow(vy0,2)-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*pow(vz0,3)+(-6*pow(vy1,4)+24*vy0*pow(vy1,3)+(-36*pow(vy0,2)-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*pow(vy1,2)+(24*pow(vy0,3)+(24*pow(vx1,2)-48*vx0*vx1+24*pow(vx0,2))*vy0)*vy1-6*pow(vy0,4)+(-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*pow(vy0,2)-6*pow(vx1,4)+24*vx0*pow(vx1,3)-36*pow(vx0,2)*pow(vx1,2)+24*pow(vx0,3)*vx1-6*pow(vx0,4))*vz0)*vz1+pow(vz0,6)+(3*pow(vy1,2)-6*vy0*vy1+3*pow(vy0,2)+3*pow(vx1,2)-6*vx0*vx1+3*pow(vx0,2))*pow(vz0,4)+(3*pow(vy1,4)-12*vy0*pow(vy1,3)+(18*pow(vy0,2)+6*pow(vx1,2)-12*vx0*vx1+6*pow(vx0,2))*pow(vy1,2)+((-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*vy0-12*pow(vy0,3))*vy1+3*pow(vy0,4)+(6*pow(vx1,2)-12*vx0*vx1+6*pow(vx0,2))*pow(vy0,2)+3*pow(vx1,4)-12*vx0*pow(vx1,3)+18*pow(vx0,2)*pow(vx1,2)-12*pow(vx0,3)*vx1+3*pow(vx0,4))*pow(vz0,2)+pow(vy1,6)-6*vy0*pow(vy1,5)+(15*pow(vy0,2)+3*pow(vx1,2)-6*vx0*vx1+3*pow(vx0,2))*pow(vy1,4)+((-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*vy0-20*pow(vy0,3))*pow(vy1,3)+(15*pow(vy0,4)+(18*pow(vx1,2)-36*vx0*vx1+18*pow(vx0,2))*pow(vy0,2)+3*pow(vx1,4)-12*vx0*pow(vx1,3)+18*pow(vx0,2)*pow(vx1,2)-12*pow(vx0,3)*vx1+3*pow(vx0,4))*pow(vy1,2)+(-6*pow(vy0,5)+(-12*pow(vx1,2)+24*vx0*vx1-12*pow(vx0,2))*pow(vy0,3)+(-6*pow(vx1,4)+24*vx0*pow(vx1,3)-36*pow(vx0,2)*pow(vx1,2)+24*pow(vx0,3)*vx1-6*pow(vx0,4))*vy0)*vy1+pow(vy0,6)+(3*pow(vx1,2)-6*vx0*vx1+3*pow(vx0,2))*pow(vy0,4)+(3*pow(vx1,4)-12*vx0*pow(vx1,3)+18*pow(vx0,2)*pow(vx1,2)-12*pow(vx0,3)*vx1+3*pow(vx0,4))*pow(vy0,2)+pow(vx1,6)-6*vx0*pow(vx1,5)+15*pow(vx0,2)*pow(vx1,4)-20*pow(vx0,3)*pow(vx1,3)+15*pow(vx0,4)*pow(vx1,2)-6*pow(vx0,5)*vx1+pow(vx0,6),-1) ,  0 ,  0 ,  0 ,  1 )); 
}

}

#endif

