#ifndef ELEMENTS_PATHSPACE_TERMINATE_HH
#define ELEMENTS_PATHSPACE_TERMINATE_HH
#include "truncPolySystem.hh"
#include "pow.hh"

namespace poly {

template<typename scalar, int numVarsIn, int numVarsOut>
static inline void terminate(truncPolySystem<scalar, numVarsIn, numVarsOut> &poly)
{
  assert(numVarsIn >= 5); // [x,y,dx,dy,f]
  poly.clear();
  // pass on almost everything
  poly[0].unity(0);
  poly[1].unity(1);
  poly[2].unity(2);
  poly[3].unity(3);

  //=================================================== 
  // taylor series degree 3 of terminated measurement contribution f 
  //=================================================== 
  // f  :  
  poly[ 4 ].add(Term4f( 1 ,  0 ,  0 ,  0 ,  0 ,  1 )); 
  // f*x^2  :  
  poly[ 4 ].add(Term4f( -42 ,  2 ,  0 ,  0 ,  0 ,  1 )); 
  // f*y^2  :  
  poly[ 4 ].add(Term4f( -42 ,  0 ,  2 ,  0 ,  0 ,  1 )); 
  // dx^2*f  :  
  poly[ 4 ].add(Term4f( -1/2 ,  0 ,  0 ,  2 ,  0 ,  1 )); 
  // dy^2*f  :  
  poly[ 4 ].add(Term4f( -1/2 ,  0 ,  0 ,  0 ,  2 ,  1 ));
}

}


#endif
