#ifndef WT_LENSFLARE_H_
#define WT_LENSFLARE_H_

#include <CL/cl.h>

typedef struct lensflare_t
{
    // store state:
    // opencl dreggn,
    // sun pos + color
    cl_context context;             // OpenCL context
    cl_command_queue command_queue; // OpenCL command queue
    cl_platform_id *platform;       // OpenCL platforms
    cl_device_id device;            // OpenCL device
    cl_program program;             // OpenCL program
    cl_kernel kernel;               // lensflare kernel
    cl_kernel kernel_image;         // transmitted image kernel

    cl_mem cl_input, cl_output;     // output buffer and input image

    char *filebuffer;               // used to optimize i/o time for large code dumps
    float sun_pos[3];               // position of the sun
    float sun_col[3];               // sun irradiance in rgb

    int verbosity;                  // print out some stats?

    void *lens;                     // buffered lens system
}
lensflare_t;

#ifdef __cplusplus
extern "C" {
#endif

// a couple of functions that should be called in this order:

// initialize a lensflare struct.
lensflare_t* lf_lensflare_init(int verbosity);

// set to non-zero to get performance stats.
void lf_lensflare_set_verbosity(lensflare_t *l, int verbosity);

// read lens description. returns the number of valid lens flares.
int lf_lensflare_setup(lensflare_t *l, const char *lensfile);
    
// setup to render using given flares and current lens description.
// set smoothing to use jacobian filtering (can be slower but smoother)
void lf_lensflare_compile(lensflare_t *l, int flare_num, int smooth);

// set the sun position and color.
// position is in mm (as the lens descriptions are)
void lf_lensflare_set_sun(lensflare_t *l, const float *pos, const float *rgb);

// render into the given buffer
// rgb float layout, should be 128-byte aligned to avoid mem copies to sse happy layout
// or better, query clGetDeviceInfo with CL_DEVICE_MEM_BASE_ADDR_ALIGN 
void lf_lensflare_render(lensflare_t *l, float *buffer, int width, int height, int num_samples);

// cleanup, free all resources. also destroys the passed pointer.
void lf_lensflare_cleanup(lensflare_t *l);

void lf_lensflare_render_image(lensflare_t *l, float *buffer, int width, int height, int num_samples, float dist, float sensordist);
void lf_lensflare_set_input_image(lensflare_t *l, int width, int height, const float *input_image);

#ifdef __cplusplus
}
#endif

#endif
