/*
 * author johannes hanika <jhanika@wetafx.co.nz>, 2012
 */
#define MM_PI 3.14156291f

const sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_LINEAR;

void splat_single(
           global float *buf,
           const float x,
           const float y,
           const float *col,
           const int width,
           const int height)
{
    if(x < width && y < height && x >= 0 && y >= 0)
        for(int c=0;c<3;c++)
            buf[3*(width*(int)y + (int)x) + c] += col[c];
}

void splat(global float *buf,
           const float x,
           const float y,
           const float *col,
           float j00, float j01, float j02, float j03,
           float j10, float j11, float j12, float j13,
           const float sample_ratio_xy,
           const float sample_ratio_uv,
           const int width,
           const int height)
{
    // simplistic jacobian sampling:
#if 0
    // only splat by dx'/{dx,dy}
    const float w[5] = {1.0f/17.0f, 4.0f/17.0f, 7.0f/17.0f, 4.0f/17.0f, 1.0f/17.0f};
    for(int i=-2;i<=2;i++)
    {
        for(int j=-2;j<=2;j++)
        {
            const float ww = w[2+i]*w[2+j];
            const float a = x + sample_ratio_xy*(j00*i + j01*j);
            const float b = y + sample_ratio_xy*(j10*i + j11*j);
            if(a < width && b < height && a >= 0 && b >= 0)
                for(int c=0;c<3;c++)
                    buf[3*(width*(int)b + (int)a) + c] += col[c] * ww;
        }
    }
#endif
    // splat by dx'/{dx,dy,du,dv}
    float p[16][5] = {
        {2.500000f,   2.500000f,  2.500000f,  2.500000f, 0.000130f},
        {2.187500f,   0.000000f,  0.833333f,  1.500000f, 0.025468f},
        {1.875000f,   1.250000f, -0.833333f,  0.500000f, 0.101871f},
        {1.562500f,  -1.250000f,  1.944444f, -0.500000f, 0.058212f},
        {1.250000f,   1.875000f,  0.277778f, -1.500000f, 0.058212f},
        {0.937500f,  -0.625000f, -1.388889f,  2.300000f, 0.025468f},
        {0.625000f,   0.625000f,  1.388889f,  1.300000f, 0.101871f},
        {0.312500f,  -1.875000f, -0.277778f,  0.300000f, 0.178274f},
        {0.000000f,   2.187500f, -1.944445f, -0.700000f, 0.025468f},
        {-0.312500f, -0.312500f,  2.314815f, -1.700000f, 0.025468f},
        {-0.625000f,  0.937500f,  0.648148f,  2.100000f, 0.044569f},
        {-0.937500f, -1.562500f, -1.018518f,  1.100000f, 0.058212f},
        {-1.250000f,  1.562500f,  1.759259f,  0.100000f, 0.058212f},
        {-1.562500f, -0.937500f,  0.092592f, -0.900000f, 0.178274f},
        {-1.875000f,  0.312500f, -1.574074f, -1.900000f, 0.058212f},
        {-2.187500f, -2.187500f,  1.203704f,  1.900000f, 0.002079f}};

    // simplistic jacobian sampling: 
    j00 *= sample_ratio_xy;
    j01 *= sample_ratio_xy;
    j10 *= sample_ratio_xy;
    j11 *= sample_ratio_xy;
    j02 *= sample_ratio_uv;
    j03 *= sample_ratio_uv;
    j12 *= sample_ratio_uv;
    j13 *= sample_ratio_uv;
    // const float w[5] = {1.0f/17.0f, 4.0f/17.0f, 7.0f/17.0f, 4.0f/17.0f, 1.0f/17.0f};
    for(int i=0;i<16;i++)
    {
        const float dx = p[i][0], dy = p[i][1], du = p[i][2], dv = p[i][3];
        const float ww = p[i][4];
        const float a = x + j00*dx + j01*dy + j02*du + j03*dv;
        const float b = y + j10*dx + j11*dy + j12*du + j13*dv;
        if(a < width && b < height && a >= 0 && b >= 0)
            for(int c=0;c<3;c++)
                buf[3*(width*(int)b + (int)a) + c] += col[c] * ww;
    }

}

/* test if a point is inside the aperture */
int
clip_aperture(const float x, const float y, const float radius)
{
    // early out
    if(x*x + y*y > radius*radius) return 0;
    const int num_blades = 8;
    float xx = radius;
    float yy = 0.0f;
    for(int b=1;b<num_blades+1;b++)
    {
        float tmpx, tmpy;
        tmpy = sincos(2.0f*(float)MM_PI/num_blades * b, &tmpx);
        tmpx *= radius;
        tmpy *= radius;
        const float normalx = xx + tmpx;
        const float normaly = yy + tmpy;
        float dot0 = (normalx)*(x-xx) + (normaly)*(y-yy);
        if(dot0 > 0.0f) return 0;
        xx = tmpx;
        yy = tmpy;
    }
    return 1;
}


constant float spectrum_s_white[] =
{
  1.0000f, 1.0000f, 0.9999f, 0.9993f, 0.9992f, 0.9998f, 1.0000f, 1.0000f, 1.0000f, 1.0000f
};
constant float spectrum_s_cyan[] =
{
  0.9710f, 0.9426f, 1.0007f, 1.0007f, 1.0007f, 1.0007f, 0.1564f, 0.0000f, 0.0000f, 0.0000f
};
constant float spectrum_s_magenta[] =
{
  1.0000f, 1.0000f, 0.9685f, 0.2229f, 0.0000f, 0.0458f, 0.8369f, 1.0000f, 1.0000f, 0.9959f 
};
constant float spectrum_s_yellow[] =
{
  0.0001f, 0.0000f, 0.1088f, 0.6651f, 1.0000f, 1.0000f, 0.9996f, 0.9586f, 0.9685f, 0.9840f 
};
constant float spectrum_s_red[] =
{
  0.1012f, 0.0515f, 0.0000f, 0.0000f, 0.0000f, 0.0000f, 0.8325f, 1.0149f, 1.0149f, 1.0149f 
};
constant float spectrum_s_green[] =
{
  0.0000f, 0.0000f, 0.0273f, 0.7937f, 1.0000f, 0.9418f, 0.1719f, 0.0000f, 0.0000f, 0.0025f 
};
constant float spectrum_s_blue[] =
{
  1.0000f, 1.0000f, 0.8916f, 0.3323f, 0.0000f, 0.0000f, 0.0003f, 0.0369f, 0.0483f, 0.0496f
};

float spectrum_rgb_to_p(const float lambda, global float *rgb)
{
  // smits-like smooth metamer construction, basis function match cie rgb backwards.
  float p = 0.0f;
  // const float corr[] = {0.467044, 0.368873, 0.351969};
  // float red = rgb[0]*corr[0], green = rgb[1]*corr[1], blue = rgb[2]*corr[2];
  float red = rgb[0], green = rgb[1], blue = rgb[2];
  float cyan = 0, yellow = 0, magenta = 0;
  const float white = min(red, min(green, blue));
  red -= white; green -= white; blue -= white;
  const int bin = (int)(10.0f*(lambda - 380.0f)/(720.0f - 380.0f));
  float ww = spectrum_s_white[bin];
  p += white * ww;
  if(green > 0.0f && blue > 0.0f)
  {
    cyan = min(green, blue);
    green -= cyan; blue -= cyan;
  }
  else if(red > 0.0f && blue > 0.0f)
  {
    magenta = min(red, blue);
    red -= magenta; blue -= magenta;
  }
  else if(red > 0.0f && green > 0.0f)
  {
    yellow = min(red, green);
    red -= yellow; green -= yellow;
  }

  float cw = spectrum_s_cyan[bin];
  float mw = spectrum_s_magenta[bin];
  float yw = spectrum_s_yellow[bin];
  p += cw*cyan;
  p += mw*magenta;
  p += yw*yellow;
  float rw = spectrum_s_red[bin];
  float gw = spectrum_s_green[bin];
  float bw = spectrum_s_blue[bin];
  p += red * rw;
  p += green * gw;
  p += blue * bw;
  return p;
}

/* tiny encryption algorithm, from holger's presentation */

float
sample_two_float(unsigned int v0, unsigned int v1, float *out)
{
    const unsigned int key0 = 0xa341316c;
    const unsigned int key1 = 0xc8013ea4;
    const unsigned int key2 = 0xad90777d;
    const unsigned int key3 = 0x7e95761e;
    unsigned int sum = 0;
    unsigned int delta = 0x9e3779b9;

    // this can be lowered for performance,
    // 16 results in reasonably good points.
    const unsigned int rounds = 16;
    for(unsigned int i = 0; i < rounds; i++)
    {
        sum += delta;
        v0 += ((v1 << 4) + key0) ^ (v1 + sum) ^ ((v1 >> 5) + key1);
        v1 += ((v0 << 4) + key2) ^ (v0 + sum) ^ ((v0 >> 5) + key3);
    }
    // convert to float in [0, 1)
    *out = v1*(1.0f/4294967296.0f);
    return v0*(1.0f/4294967296.0f);
}

kernel void render(
    global float        *output,
    global float        *input,
    const float          dist,
    const float          pupil_radius,
    const int            num_samples,
    const int            width,
    const int            height,
    const float          sensordist)
{
    // grid positions are random seeds:
    const int r1 = get_global_id(0);
    const int r2 = get_global_id(1);

    const float sample_ratio_xy = native_rsqrt((float)num_samples) * pupil_radius * width/36.0f;

    // draw a few samples
    for(int k=0;k<num_samples;k++)
    {
        float rnd0, rnd1; // position on incoming pupil
        rnd0 = sample_two_float(r1+k, r2+height*k, &rnd1);
        float rnd2, rnd3; // anti-alias
        rnd2 = sample_two_float(r1+1337+k, r2+height*k, &rnd3);
        float rnd4, rnd5; // lambda
        rnd4 = sample_two_float(r1+7331+k, r2+height*k, &rnd5);

        // get pixel position from input image, antialised.
        // no need to clip because we use an image sampler.
        const float i = rnd2 + r1;
        const float j = rnd3 + r2;

        const float lambda = 400.0f + 300.0f*rnd4;
        const float L = spectrum_rgb_to_p(lambda, input + 3*(width*r2 + r1))/num_samples;

        // sample incoming pupil (round)
        float tmpx, tmpy;
        tmpy = sincos(2.0f*MM_PI*rnd0, &tmpx);
        const float r = native_sqrt(rnd1)*pupil_radius;
        const float x = r*tmpx;
        const float y = r*tmpy;

        // fov: atan(0.3/1.0)
        float dx = (x - dist * 0.25f*(i-width *.5f)/width);
        float dy = (y - dist * 0.25f*(j-height*.5f)/width);
        // dz == dist
        const float normr = 1.0f/dist;
        dx *= normr;
        dy *= normr;

        const float sample_ratio_uv = sample_ratio_xy * (dist * normr) * normr * normr;

        const float scale = (float)width/36.0f;

        // begin autogenerated part:

