#ifndef POLY_PRINT_HH
#define POLY_PRINT_HH

#include "truncPolySystem.hh"

namespace poly {

/* Non-member functions for I/O */


_tSN
void print(std::ostream& os, const polyTerm _SN term, std::string *varNames = NULL) {

  os << std::setprecision(16) << term.coefficient;

  if (varNames) {
    for(int i=0; i<numVars; i++) {
      echar e = term.exponents[i];
      std::string vn = varNames[i];
      if (e)
	os << " " <<vn;
      if (e>1)
	os << "^" << int(e);
    }
    
  } else {
    for(int i=0; i<numVars; i++) {
      echar e = term.exponents[i];
      if (e)
	os << " * x" << i;
      if (e>1)
	os << "^" << int(e);
    }

  }
}

_tSN
std::ostream& operator<<(std::ostream& os, const polyTerm _SN& term) {

  print(os, term);
  return os;
}

_tSN
void print(std::ostream& os, const truncPoly _SN poly, std::string *varNames = NULL, bool linebreak = false) {

  int N = poly.terms.size();

  if (N) {
    for(int i = 0; i < N - 1; i++) {
      print(os, poly.terms[i], varNames);
      if (linebreak) os << std::endl;
      os << " + ";
    }
    
    print(os, poly.terms[N-1], varNames);
   
  } else os << "[Null]";

}

_tSN
std::ostream& operator<<(std::ostream& os, const truncPoly _SN& poly) {

  print(os, poly, NULL, false);      
  return os;
}

_tSIO
void print(std::ostream& os, const truncPolySystem _SIO& tpt, std::string *varNames = NULL, bool linebreak = false)
{
  for(int i = 0; i < numVarsOut; i++) {
    if (varNames) { 
      std::string vn = varNames[i];
      os << vn << "' = ";
    } else {
      os << "x" << i << "' = ";
    }

    print(os,tpt.equations[i],varNames,linebreak);
    os << std::endl;
  }
}


_tSIO
std::ostream& operator<<(std::ostream& os, const truncPolySystem _SIO& tpt) {

  print(os, tpt);
  return os;
}

} // namespace

#endif
