
#include "lensflare.h"

#include <stdio.h>
#include <stdlib.h>

int main(int argc, char *argv[])
{
    lensflare_t *flare = lf_lensflare_init(1);

    int width = 1024, height = 1024, num_samples = 100;
    float *buffer = 0;
    if(posix_memalign((void **)&buffer, 1024, width*height*3*sizeof(float)))
    {
        fprintf(stderr, "[lensflare_test] failed to allocate buffer.\n");
        exit(1);
    }

    // see lenses/ for more input data:
    // int num_flares = lf_lensflare_setup(flare, "lenses/Canon-MassiveAnamorphicZoom.txt");
    // int num_flares = lf_lensflare_setup(flare, "lenses/lensbaby.fx");
    // int num_flares = lf_lensflare_setup(flare, "lenses/ItohZoom.fx");
    lf_lensflare_setup(flare, "lenses/petzval.fx");
    // -1 means all possible flares (two-bounce *R*R* type paths only)
    // -2 means directly transmitted light (to form an image)
    // second argument is smoothing on (!=0) or off (0)
    lf_lensflare_compile(flare, -2, 1);//num_flares/2, 0);
    const float rgb[3] = {1., 1., 1.};
    const float sun[3] = {100., 0., 1000.};
    lf_lensflare_set_sun(flare, sun, rgb);

    float *input_image = (float *)malloc(width*height*sizeof(float)*3);
    {
        FILE *f = fopen("test.pfm", "rb");
        if(!f)
        {
            free(input_image);
            input_image = NULL;
        }
        else
        {
            int wd, ht;
            int err = fscanf(f, "PF\n%d %d\n%*[^\n]", &wd, &ht);
            err = fgetc(f); // eat \n, but only one of them.
            err += fread(input_image, sizeof(float), 3*width*height, f);
            fclose(f);
            fprintf(stderr, "[image-test] loaded input.pfm with %dx%d (%d)\n", wd, ht, err);
        }
    }
    lf_lensflare_set_input_image(flare, width, height, input_image);
    // iterate over 0..33..100 to get a nice zoom
    for(int k=0;k<100;k++)
    {
      lf_lensflare_render_image(flare, buffer, width, height, num_samples, 1000000+k*20.f, 0.0f);//130.0f);
      char filename[512];
      snprintf(filename, 512, "img_%04d.pfm", k);
      // write out what we've got:
      FILE *f = fopen(filename, "wb");
      fprintf(f, "PF\n%d %d\n-1.0\n", width, height);
      fwrite(buffer, width*height*3, sizeof(float), f);
      fclose(f);
    }

    lf_lensflare_cleanup(flare);

    free(buffer);
    exit(0);
}
